# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #
import numpy as np
import pytest

import xtrack as xt
import xobjects as xo
import xpart as xp


def test_check_is_active_sorting_openmp():
    test_context = xo.ContextCpu(omp_num_threads=5)

    class TestElement(xt.BeamElement):
        _xofields = {
            'states': xo.Int64[:],
        }

        _extra_c_sources = ["""
            #define XT_OMP_SKIP_REORGANIZE

            /*gpufun*/
            void TestElement_track_local_particle(
                TestElementData el,
                LocalParticle* part0
            ) {
                //start_per_particle_block (part0->part)
                    int64_t state = check_is_active(part);
                    int64_t id = LocalParticle_get_particle_id(part);
                    TestElementData_set_states(el, id, state);
                //end_per_particle_block
            }
        """]

    el = TestElement(
        _context=test_context,
        states=np.zeros(18, dtype=np.int64),
    )
    particles = xp.Particles(
        _context=test_context,
        state=[
            1, 0, 1, 0, 1,  # should be reordered to 1, 1, 1, 0, 0
            0, 0, 0, 0, 0,  # should be left intact
            0, 1, 0, 1, 1,  # should be reordered to 1, 1, 1, 0, 0
            1, 1, 0,        # should be left intact
        ],
        _capacity=22,       # there are 4 particles that are unallocated
        _no_reorganize=True,
    )

    el.track(particles)

    # We have five threads, so the particles should be split into chunks
    # of 5, 5, 5, 3 + 2 (unallocated), 2 (unallocated).
    assert len(particles.state) == 22

    # Check that each chunk is reorganized correctly.
    # First batch:
    assert np.all(particles.state[0:5] == [1, 1, 1, 0, 0])
    assert set(particles.particle_id[0:3]) == {0, 2, 4}
    assert set(particles.particle_id[3:5]) == {1, 3}

    # Second batch:
    assert np.all(particles.state[5:10] == [0, 0, 0, 0, 0])
    # Don't reorder if not needed:
    assert np.all(particles.particle_id[5:10] == [5, 6, 7, 8, 9])

    # Third batch:
    assert np.all(particles.state[10:15] == [1, 1, 1, 0, 0])
    assert set(particles.particle_id[10:13]) == {11, 13, 14}
    assert set(particles.particle_id[13:15]) == {10, 12}

    # Fourth batch:
    assert np.all(particles.state[15:20] == [1, 1, 0, -999999999, -999999999])
    # Don't reorder if not needed:
    assert np.all(particles.particle_id[15:18] == [15, 16, 17])

    # Fifth batch (unallocated):
    assert np.all(particles.state[20:22] == [-999999999, -999999999])


@pytest.mark.parametrize(
    'test_context',
    [
        xo.ContextCpu(),
        xo.ContextCpu(omp_num_threads=4),
    ]
)
def test_check_is_active_sorting_cpu_default(test_context):
    class TestElement(xt.BeamElement):
        _xofields = {
            'states': xo.Int64[:],
        }

        _extra_c_sources = ["""
            /*gpufun*/
            void TestElement_track_local_particle(
                TestElementData el,
                LocalParticle* part0
            ) {
                //start_per_particle_block (part0->part)
                    int64_t state = check_is_active(part);
                    int64_t id = LocalParticle_get_particle_id(part);
                    TestElementData_set_states(el, id, state);
                //end_per_particle_block
            }
        """]

    el = TestElement(
        _context=test_context,
        states=np.zeros(18, dtype=np.int64),
    )
    particles = xp.Particles(
        _context=test_context,
        state=[
            1, 0, 1, 0, 1,
            0, 0, 0, 0, 0,
            0, 1, 0, 1, 1,
            1, 1, 0,
        ],
        _no_reorganize=True,
    )
    # We want to simulate a situation where a recount is needed, so we put a
    # value of active particles to be equal to the total number of particles:
    particles._num_active_particles = 18

    el.track(particles)

    # Here we don't reorganize by batches, so we just check the whole array
    # to see if it's sensible:
    assert np.all(particles.state == ([1] * 8) + ([0] * 10))
    assert set(particles.particle_id[:8]) == {0, 2, 4, 11, 13, 14, 15, 16}
    assert set(particles.particle_id[8:]) == {1, 3, 5, 6, 7, 8, 9, 10, 12, 17}



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp
from xpart.pdg import _PDG, _elements, _elements_long, _pdg_id_ion, _mass_consistent

from xobjects.test_helpers import for_all_test_contexts
import xobjects as xo

def test_names():
    names = [val[1] for val in _PDG.values()]
    names += [f"{val}174" for val in _elements.values()]
    names += [f"{val} 134" for val in _elements_long.values()]
    for name in [val[1] for val in _PDG.values()]:
        pdg_id = xp.pdg.get_pdg_id_from_name(name)
        assert xp.pdg.get_name_from_pdg_id(pdg_id) == name
        assert xp.pdg.get_pdg_id_from_name(
            xp.pdg.get_name_from_pdg_id(pdg_id)) == pdg_id


def test_lead_208():
    pdg_id = 1000822080
    assert _pdg_id_ion(208, 82) == pdg_id
    assert xp.pdg.get_pdg_id_from_mass_charge(xp.Pb208_MASS_EV, 82) == pdg_id
    assert xp.pdg.get_name_from_pdg_id(pdg_id) == 'Pb208'
    assert xp.pdg.get_pdg_id_from_name('Pb208')    == pdg_id
    assert xp.pdg.get_pdg_id_from_name('Pb 208')   == pdg_id
    assert xp.pdg.get_pdg_id_from_name('Pb-208')   == pdg_id
    assert xp.pdg.get_pdg_id_from_name('Pb_208')   == pdg_id
    assert xp.pdg.get_pdg_id_from_name('Pb.208')   == pdg_id
    assert xp.pdg.get_pdg_id_from_name('pb208')    == pdg_id
    assert xp.pdg.get_pdg_id_from_name('lead208')  == pdg_id
    assert xp.pdg.get_pdg_id_from_name('Lead 208') == pdg_id
    assert xp.pdg.get_pdg_id_from_name('Lead_208') == pdg_id
    assert _mass_consistent(pdg_id, xp.Pb208_MASS_EV)
    assert xp.pdg.get_element_name_from_Z(82) == 'Pb'
    assert xp.pdg.get_element_full_name_from_Z(82) == 'Lead'
    assert np.allclose(xp.pdg.get_mass_from_pdg_id(pdg_id), xp.Pb208_MASS_EV,
                       rtol=1e-10, atol=0)
    assert xp.pdg.get_properties_from_pdg_id(pdg_id) == (82., 208, 82, 'Pb208')

@for_all_test_contexts
def test_build_reference_from_pdg_id(test_context):
    particle_ref_proton  = xp.reference_from_pdg_id(pdg_id='proton',
                                                    _context=test_context)
    particle_ref_proton.move(_context=xo.context_default)
    assert particle_ref_proton.pdg_id == 2212
    particle_ref_lead = xp.reference_from_pdg_id(pdg_id='Pb208',
                                                 _context=test_context)
    particle_ref_lead.move(_context=xo.context_default)
    assert np.allclose(particle_ref_lead.q0, 82.)
    assert np.allclose(particle_ref_lead.mass0, xp.Pb208_MASS_EV)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import numpy as np
from numbers import Number

from xtrack.particles.constants import U_MASS_EV, PROTON_MASS_EV, ELECTRON_MASS_EV, MUON_MASS_EV, Pb208_MASS_EV


# Monte Carlo numbering scheme as defined by the Particle Data Group
# See https://pdg.lbl.gov/2007/reviews/montecarlorpp.pdf for implementation
# details.
# Not all particles are implemented yet; this can be appended later


_PDG = {
#   ID       q  NAME
    0:     [0.,  'undefined'],
    11:    [-1., 'electron'],
    -11:   [1.,  'positron'],
    13:    [-1., 'muon'],
    15:    [-1., 'tau'],
    111:   [0.,  'pi0'],
    211:   [1.,  'pi+'],
    -211:  [-1., 'pi-'],
    311:   [0.,  'K0'],
    321:   [1.,  'K+'],
    -321:  [-1., 'K-'],
    2212:  [1.,  'proton'],
    2112:  [0.,  'neutron'],
    2224:  [2.,  'Delta++'],
    2214:  [1.,  'Delta+'],
    2114:  [0.,  'Delta0'],
    1114:  [-1., 'Delta-'],
    3122:  [0.,  'Lambda'],
    3222:  [1.,  'Sigma+'],
    3212:  [0.,  'Sigma0'],
    3112:  [-1., 'Sigma-'],
    3322:  [0.,  'Xi'],
    3312:  [-1., 'Xi-'],
    1000010020: [1., 'deuterium'],
    1000010030: [1., 'tritium']
}

_elements = {
     1: "H",    2: "He",   3: "Li",   4: "Be",   5: "B",    6: "C",    7: "N",    8: "O",    9: "F",   10: "Ne",
    11: "Na",  12: "Mg",  13: "Al",  14: "Si",  15: "P",   16: "S",   17: "Cl",  18: "Ar",  19: "K",   20: "Ca",
    21: "Sc",  22: "Ti",  23: "V",   24: "Cr",  25: "Mn",  26: "Fe",  27: "Co",  28: "Ni",  29: "Cu",  30: "Zn",
    31: "Ga",  32: "Ge",  33: "As",  34: "Se",  35: "Br",  36: "Kr",  37: "Rb",  38: "Sr",  39: "Y",   40: "Zr",
    41: "Nb",  42: "Mo",  43: "Tc",  44: "Ru",  45: "Rh",  46: "Pd",  47: "Ag",  48: "Cd",  49: "In",  50: "Sn",
    51: "Sb",  52: "Te",  53: "I",   54: "Xe",  55: "Cs",  56: "Ba",  57: "La",  58: "Ce",  59: "Pr",  60: "Nd",
    61: "Pm",  62: "Sm",  63: "Eu",  64: "Gd",  65: "Tb",  66: "Dy",  67: "Ho",  68: "Er",  69: "Tm",  70: "Yb",
    71: "Lu",  72: "Hf",  73: "Ta",  74: "W",   75: "Re",  76: "Os",  77: "Ir",  78: "Pt",  79: "Au",  80: "Hg",
    81: "Tl",  82: "Pb",  83: "Bi",  84: "Po",  85: "At",  86: "Rn",  87: "Fr",  88: "Ra",  89: "Ac",  90: "Th",
    91: "Pa",  92: "U",   93: "Np",  94: "Pu",  95: "Am",  96: "Cm",  97: "Bk",  98: "Cf",  99: "Es", 100: "Fm",
   101: "Md", 102: "No", 103: "Lr", 104: "Rf", 105: "Db", 106: "Sg", 107: "Bh", 108: "Hs", 109: "Mt", 110: "Ds",
   111: "Rg", 112: "Cn", 113: "Nh", 114: "Fl", 115: "Mc", 116: "Lv", 117: "Ts", 118: "Og"
}

_elements_long = {
    1: "Hydrogen",        2: "Helium",          3: "Lithium",         4: "Beryllium",       5: "Boron",
    6: "Carbon",          7: "Nitrogen",        8: "Oxygen",          9: "Fluorine",       10: "Neon",
   11: "Sodium",         12: "Magnesium",      13: "Aluminum",       14: "Silicon",        15: "Phosphorus",
   16: "Sulfur",         17: "Chlorine",       18: "Argon",          19: "Potassium",      20: "Calcium",
   21: "Scandium",       22: "Titanium",       23: "Vanadium",       24: "Chromium",       25: "Manganese",
   26: "Iron",           27: "Cobalt",         28: "Nickel",         29: "Copper",         30: "Zinc",
   31: "Gallium",        32: "Germanium",      33: "Arsenic",        34: "Selenium",       35: "Bromine",
   36: "Krypton",        37: "Rubidium",       38: "Strontium",      39: "Yttrium",        40: "Zirconium",
   41: "Niobium",        42: "Molybdenum",     43: "Technetium",     44: "Ruthenium",      45: "Rhodium",
   46: "Palladium",      47: "Silver",         48: "Cadmium",        49: "Indium",         50: "Tin",
   51: "Antimony",       52: "Tellurium",      53: "Iodine",         54: "Xenon",          55: "Cesium",
   56: "Barium",         57: "Lanthanum",      58: "Cerium",         59: "Praseodymium",   60: "Neodymium",
   61: "Promethium",     62: "Samarium",       63: "Europium",       64: "Gadolinium",     65: "Terbium",
   66: "Dysprosium",     67: "Holmium",        68: "Erbium",         69: "Thulium",        70: "Ytterbium",
   71: "Lutetium",       72: "Hafnium",        73: "Tantalum",       74: "Tungsten",       75: "Rhenium",
   76: "Osmium",         77: "Iridium",        78: "Platinum",       79: "Gold",           80: "Mercury",
   81: "Thallium",       82: "Lead",           83: "Bismuth",        84: "Polonium",       85: "Astatine",
   86: "Radon",          87: "Francium",       88: "Radium",         89: "Actinium",       90: "Thorium",
   91: "Protactinium",   92: "Uranium",        93: "Neptunium",      94: "Plutonium",      95: "Americium",
   96: "Curium",         97: "Berkelium",      98: "Californium",    99: "Einsteinium",   100: "Fermium",
  101: "Mendelevium",   102: "Nobelium",      103: "Lawrencium",    104: "Rutherfordium", 105: "Dubnium",
  106: "Seaborgium",    107: "Bohrium",       108: "Hassium",       109: "Meitnerium",    110: "Darmstadtium",
  111: "Roentgenium",   112: "Copernicium",   113: "Nihonium",      114: "Flerovium",     115: "Moscovium",
  116: "Livermorium",   117: "Tennessine",    118: "Oganesson"
}

def get_name_from_pdg_id(pdg_id):
    if hasattr(pdg_id, '__len__'):
        return np.array([get_name_from_pdg_id(pdg) for pdg in pdg_id])
    return get_properties_from_pdg_id(pdg_id)[-1]


def get_pdg_id_from_name(name=None):

    if hasattr(name, 'get'):
        name = name.get()

    if name is None:
        return 0  # undefined
    elif hasattr(name, '__len__') and not isinstance(name, str):
        return np.array([get_pdg_id_from_name(nn) for nn in name])
    elif isinstance(name, Number):
        return int(name) # fallback

    _PDG_inv      = {val[1].lower(): pdg_id for pdg_id, val in _PDG.items()}

    lname = name.lower()
    aname = ""
    if len(lname) > 4 and lname[:5]=="anti-":
        aname = lname[5:]

    # particle
    if lname in _PDG_inv.keys():
        return _PDG_inv[lname]

    # anti-particle
    elif aname in _PDG_inv.keys():
        return -_PDG_inv[aname]

    else:
        ion_long = [[Z, ion.lower()] for Z, ion in _elements_long.items()
                    if lname.startswith(ion.lower())]
        ion      = [[Z, ion.lower()] for Z, ion in _elements.items()
                    if lname.startswith(ion.lower())]

        # ion short name
        if len(ion_long) > 0:
            # Multiple finds are possible (e.g. [15, P] and [82, Pb] in the case of lead).
            # Use the result that has the longest name length (most restrictive result)
            ion_long.sort(key=lambda el: len(el[1]))
            ion_long = ion_long[-1]
            Z = ion_long[0]
            try:
                A = int(lname.replace(ion_long[1], '').replace('.','').replace('_','').replace('-','').replace(' ',''))
            except:
                raise ValueError(f"Wrongly formatted ion name: cannot deduce A from {name}!\n"
                               + f"Use e.g. 'Pb208', 'Pb 208', 'Pb-208', 'Pb_208', or 'Pb.208'.")
            return _pdg_id_ion(A, Z)

        # ion short name
        elif len(ion) > 0:
            # Multiple finds are possible (e.g. [15, P] and [82, Pb] in the case of lead).
            # Use the result that has the longest name length (most restrictive result)
            ion.sort(key=lambda el: len(el[1]))
            ion = ion[-1]
            Z = ion[0]
            try:
                A = int(lname.replace(ion[1], '').replace('.','').replace('_','').replace('-','').replace(' ',''))
            except:
                raise ValueError(f"Wrongly formatted ion name: cannot deduce A from {name}!\n"
                               + f"Use e.g. 'Pb208', 'Pb 208', 'Pb-208', 'Pb_208', or 'Pb.208'.")
            return _pdg_id_ion(A, Z)

        else:
            raise ValueError(f"Particle {name} not found in pdg dictionary!")


# TODO: mass info ?
# q, A, Z, name
def get_properties_from_pdg_id(pdg_id):
    if hasattr(pdg_id, '__len__'):
        result = np.array([get_properties_from_pdg_id(pdg) for pdg in pdg_id]).T
        return result[0].astype(np.float64), result[1].astype(np.int64),\
               result[2].astype(np.int64), result[3]

    if pdg_id in _PDG.keys():
        q = _PDG[pdg_id][0]
        name = _PDG[pdg_id][1]
        if name=='proton' or name=='neutron':
            A = 1
            Z = q
        elif name=='deuterium':
            A = 2
            Z = 1
        elif name=='tritium':
            A = 3
            Z = 1
        else:
            A = 0
            Z = 0
        return float(q), int(A), int(Z), name
    elif -pdg_id in _PDG.keys():
        antipart = get_properties_from_pdg_id(-pdg_id)
        return -antipart[0], antipart[1], -antipart[2], f'anti-{antipart[3]}'

    elif pdg_id > 1000000000:
        # Ion
        tmpid = pdg_id - 1000000000
        L = int(tmpid/1e7)
        tmpid -= L*1e7
        Z = int(tmpid /1e4)
        tmpid -= Z*1e4
        A = int(tmpid /10)
        tmpid -= A*10
        isomer_level = int(tmpid)
        return float(Z), int(A), int(Z), f'{get_element_name_from_Z(Z)}{A}'#, L, isomer_level, get_element_full_name_from_Z(Z)

    else:
        raise ValueError(f"PDG ID {pdg_id} not recognised!")


def get_element_name_from_Z(z):
    if z not in _elements:
        raise ValueError(f"Element with {z} protons not known.")
    return _elements[z]


def get_element_full_name_from_Z(z):
    if z not in _elements_long:
        raise ValueError(f"Element with {z} protons not known.")
    return _elements_long[z]


def _pdg_id_ion(A, Z):
    if hasattr(A, '__len__') and hasattr(Z, '__len__'):
        return np.array([_pdg_id_ion(aa, zz) for aa, zz in zip(A,Z)])
    return  int(1000000000 + Z*10000 + A*10)


# TODO: this should be done a bit nicer, with a lookup table for the masses with A = 0
def get_pdg_id_from_mass_charge(m, q):
    if hasattr(q, '__len__') and hasattr(m, '__len__'):
        return np.array([get_pdg_id_from_mass_charge(mm, qq) for qq, mm in zip(q, m)])
    elif hasattr(q, '__len__'):
        return np.array([get_pdg_id_from_mass_charge(m, qq) for qq in q])
    elif hasattr(m, '__len__'):
        return np.array([get_pdg_id_from_mass_charge(mm, q) for mm in m])

    A = round(m/U_MASS_EV)
    if abs(m-ELECTRON_MASS_EV) < 100:
        return get_pdg_id_from_name('electron')
    elif abs(m-MUON_MASS_EV) < 100:
        return get_pdg_id_from_name('muon')
    elif abs(m-PROTON_MASS_EV) < 1000:
        return get_pdg_id_from_name('proton')
    elif q <= 0 or A <= 0:
        raise ValueError(f"Particle with {q=} and {m=} not recognised!")
    else:
        return _pdg_id_ion(A, q)


def _mass_consistent(pdg_id, m, mask=None):
    if hasattr(pdg_id, '__len__') and hasattr(m, '__len__'):
        if mask is None:
            return np.all([_mass_consistent(pdg, mm) for pdg, mm in zip(pdg_id, m)])
        else:
            pdg_id = np.array(pdg_id)[mask]
            m = np.array(m)[mask]
            return np.all([_mass_consistent(pdg, mm) for pdg, mm in zip(pdg_id, m)])
    elif hasattr(pdg_id, '__len__'):
        if mask is None:
            return np.all([_mass_consistent(pdg, m) for pdg in pdg_id])
        else:
            pdg_id = np.array(pdg_id)[mask]
            return np.all([_mass_consistent(pdg, m) for pdg in pdg_id])
    elif hasattr(m, '__len__'):
        if mask is None:
            return np.all([_mass_consistent(pdg_id, mm) for mm in m])
        else:
            m = np.array(m)[mask]
            return np.all([_mass_consistent(pdg_id, mm) for mm in m])

    q, A, _, name = get_properties_from_pdg_id(pdg_id)
    if name=='proton' or name=='electron' or name=='muon':
        return pdg_id == get_pdg_id_from_mass_charge(m, q)
    elif A > 1:
        return A==round(m/U_MASS_EV)
    else:
        # No check for other particles
        return True


# TODO: this should be done a bit nicer, with a lookup table
def get_mass_from_pdg_id(pdg_id, allow_approximation=True, expected_mass=None):
    if hasattr(pdg_id, '__len__'):
        return np.array([get_mass_from_pdg_id(pdg,
                                      allow_approximation=allow_approximation,
                                      expected_mass=expected_mass)
                         for pdg in pdg_id])

    _, A, _, name = get_properties_from_pdg_id(pdg_id)
    if name == 'proton':
        return PROTON_MASS_EV
    elif name == 'electron':
        return ELECTRON_MASS_EV
    elif name == 'muon':
        return MUON_MASS_EV
    elif name == 'Pb208':
        return Pb208_MASS_EV
    elif allow_approximation and A>0:
        print(f"Warning: approximating the mass as {A}u!")
        return A*U_MASS_EV
    else:
        if expected_mass is not None and _mass_consistent(pdg_id, expected_mass):
            # This is a workaround in case an exact mass is given
            # (like for the reference particle)
            return expected_mass
        else:
            raise ValueError(f"Exact mass for {name} not found.")





